# -*- coding: UTF-8 -*-
import Queue
import atexit
import fnmatch
import json
import multiprocessing
import os
import signal
import startReader
import tempfile
import threading
import time
import traceback

from common import logger, config
from artix.common import servicelock
from unload.service.reader import Reader
from prettytable import PrettyTable, ALL
from httplib import INTERNAL_SERVER_ERROR
from bottle import get, post, response, request, Bottle, run

log = logger.getUnloadLogger('main')


def task_unload(shop, work_directory, task, cash):
    if multiprocessing.current_process().name == 'MainProcess':
        log.debug("Старт выгрузки продаж для узла:  магазин: %s - касса %s" % (shop.id, cash.cashId))
    else:
        log.debug("Поток [%s] начал выгрузку продаж для узла:  магазин: %s - касса %s" % (
            multiprocessing.current_process().name, shop.id, cash.cashId))
    result = None
    if shop and task:
        result = startReader.unload(shop, work_directory, shifts=task.get('shifts'),
                                    dates=task.get('dates'),
                                    granularity=task.get('granularity', 'document'),
                                    closedShifts=task.get('closedShifts', 0),
                                    changedShifts=task.get('changedShifts'),
                                    cashCode=cash.code)
    if multiprocessing.current_process().name == 'MainProcess':
        log.debug("Завершена выгрузка продаж для узла:  магазин: %s - касса %s" % (shop.id, cash.cashId))
    else:
        log.debug("Поток [%s] закончил выгрузку продаж для узла:  магазин: %s - касса %s, результат: %s " % (
            multiprocessing.current_process().name, shop.id, cash.cashId if cash else None, result))

    return result, cash.cashId if cash else None


class UnloadService(Bottle):

    def __init__(self):
        super(UnloadService, self).__init__()
        get("/exchangers/sales/unload/queue")(self.__get_unload_tasks_queue_status)
        post("/exchangers/sales/unload")(self.__add_unload_task_to_queue)
        self.__readConfig()
        self.nodes = []
        self.lockFile = os.path.normpath('%s/locks/lockArtixUnloadService' % tempfile.gettempdir())
        self.srvLock = servicelock.ServiceLock(self.lockFile, log)
        self.reader = Reader()
        self.unloadTasks = Queue.PriorityQueue(5000)
        atexit.register(self.srvLock.unlock)
        signal.signal(signal.SIGTERM, self.srvLock.unlock)
        self.timeoutNodeReread = 300  # таймаут перечитывания настроек
        self.notProcessedShops = []
        self.unloadTask = None
        self.taskCountPass = 0
        self.not_processed_cashes = []
        self.all_cashes = 0
        self.processed_cashes = 0
        self.current_cash = None
        self.unloadTaskPriority = 10

    def __start(self):
        log.info('Запуск REST-сервиса выгрузки продаж в АСТУ, host = %s, port = %s' % (
            self.restServiceHost, self.restServicePort))
        try:
            if not self.srvLock.lock(False):
                log.error("Сервис уже запущен")
                exit(1)  # выход без выполнения atexit
            t = threading.Thread(target=run,
                                 kwargs={'host': self.restServiceHost, 'port': self.restServicePort, 'debug': True})
            t.daemon = True
            t.start()
        except Exception as e:
            log.error(u'Ошибка: %s' % e)
            log.error(traceback.format_exc())

    def __readConfig(self):
        self.unloadCurrentSalesInterval = config.getInt('UNLOAD', 'unloadCurrentSalesInterval', 600)
        self.unloadSalesInterval = config.getInt('UNLOAD', 'unloadSalesInterval', 20)
        self.closedShiftsByFileFlag = config.getInt('UNLOAD', 'closedShiftsByFileFlag', 2)
        self.workDirectory = config.getString('main', 'workDirectory',
                                              '%s/artixExchange/workDir' % tempfile.gettempdir())
        self.multi_processing = config.getBool('UNLOAD', 'multiprocessing', False)
        self.processes_in_pool_count = config.getInt('UNLOAD', 'processCount', 10)
        self.restServiceHost = config.getString('UNLOAD', 'restHost', '0.0.0.0')
        self.restServicePort = config.getString('UNLOAD', 'restPort', 28082)

        log.debug('main.workDirectory - %s' % self.workDirectory)
        log.debug('UNLOAD.unloadCurrentSalesInterval - %s' % self.unloadCurrentSalesInterval)
        log.debug('UNLOAD.unloadSalesInterval - %s' % self.unloadSalesInterval)
        log.debug('UNLOAD.closedShiftsByFileFlag - %s' % self.closedShiftsByFileFlag)
        log.debug('UNLOAD.multiprocessing - %s' % self.multi_processing)
        log.debug('UNLOAD.processesInPool - %s' % self.processes_in_pool_count)

    def on_start(self):
        unloadCurrentSalesInterval = self.unloadCurrentSalesInterval
        unloadSalesInterval = self.unloadSalesInterval
        closedShiftsByFileFlag = self.closedShiftsByFileFlag
        workDirectory = self.workDirectory
        timeNodeReread = 0
        timeUnloadCurrentSales = 0
        timeFindRequest = 0
        nodeGetter = None

        try:
            nodeGetter = __import__('unload.node.%s.nodeGetter' % config.getString('UNLOAD', 'nodeGetter', 'cs30'),
                                    fromlist='q')
            self.nodes = nodeGetter.getNodes()
            timeNodeReread = time.time()
        except Exception as e:
            log.error(u'Ошибка получения нод. %s' % str(e))
            log.error(traceback.format_exc())
        self.__start()
        passCountSending = 0
        while True:
            if (time.time() - self.timeoutNodeReread) > timeNodeReread:
                timeNodeReread = time.time()
                try:
                    newNodes = nodeGetter.getNodes()
                    self.nodes = newNodes
                except:
                    pass

            if unloadCurrentSalesInterval != 0:
                if (time.time() - unloadCurrentSalesInterval) > timeUnloadCurrentSales:
                    log.info(u'Выгрузка текущих продаж...')
                    timeUnloadCurrentSales = time.time()
                    for node in self.nodes:
                        for cash in node.cashes:
                            for converter in cash.converters:
                                if hasattr(converter, 'onlineSalePath') and converter.onlineSalePath:
                                    log.info('Unload current sales: shopId=%s, cashCode=%s, onlineSalePath=%s' % (
                                        node.id, cash.code, converter.onlineSalePath))
                                    startReader.unload(node, workDirectory, converterName=converter.name,
                                                       cashCode=cash.code, closedShifts=1)
            #### Выгрузка по таску из REST'а
            self.__unload_by_task(self.nodes)
            ##################################################################
            if timeFindRequest == 0 or (time.time() - unloadSalesInterval) > timeFindRequest:
                timeFindRequest = time.time()

                for node in self.nodes:
                    for cash in node.cashes:
                        for converter in cash.converters:
                            if converter.flag:
                                requestFile = self.__findFile(os.path.dirname(converter.flag),
                                                              os.path.basename(converter.flag))
                                if requestFile:
                                    log.info(u'Выгрузка продаж по запросу файла-флага %s' % converter.flag)
                                    startReader.unload(node, workDirectory, request=requestFile,
                                                       converterName=converter.name, cashCode=cash.code,
                                                       closedShifts=closedShiftsByFileFlag)
            if passCountSending < 5:
                passCountSending += 1
            else:
                passCountSending = 0
                for node in self.nodes:
                    if self.__findFile(workDirectory + '/%s/unload' % node.id, 'ready'):
                        self.reader.send(node, workDirectory)
                    if self.__findFile(workDirectory + '/%s/unload' % node.id, 'readyOnline'):
                        self.reader.send(node, workDirectory, closedShifts=1)
            time.sleep(2)

    @staticmethod
    def __findFile(directory, pattern):
        if os.path.exists(directory):
            for file in os.listdir(directory):
                if fnmatch.fnmatch(file.lower(), pattern.lower()):
                    return os.path.join(directory, file)
        return None

    def __unload_by_task(self, nodes):
        # пытаемся выполнить задачу на выгрузку не больше 20 раз, затем пропускаем ее
        self.processed_cashes = 0
        self.current_cash = None
        if self.taskCountPass > 20:
            log.warning('Невозможно выполнить задачу %s для магазинов %s' % (self.unloadTask, self.not_processed_cashes))
            log.warning('Кол-во попыток выгрузки истекло. Приступаем к следующей задаче...')
            del self.not_processed_cashes[:]
            self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)

        if len(self.not_processed_cashes) == 0:
            self.taskCountPass = 0
            try:
                if self.unloadTasks.unfinished_tasks > 0:
                    self.unloadTaskPriority, self.unloadTask = self.unloadTasks.queue.__getitem__(0)
                    log.info('Сервис начал выполнять задачу %s' % self.unloadTask)
                    # собираем магазины из задачи или берем все
                    self.notProcessedShops = [str(self.unloadTask.get('shopId'))] if (self.unloadTask.get('shopId')) else [str(id) for id in self.unloadTask.get('shopIds')] if (self.unloadTask.get('shopIds')) else [str(n.id) for n in nodes]
                    log.info('Найдены магазины: %s' % self.notProcessedShops)
                    # собираем кассы для выгрузки по отобранным магазинам
                    for shop in filter(lambda x: x.id in self.notProcessedShops, nodes):
                        for cash in shop.cashes:
                            # случай выгрузки в групповых операциях по кассам
                            if self.unloadTask.get('cashIds') and cash.cashId in self.unloadTask.get('cashIds'):
                                self.not_processed_cashes.append(cash.cashId)
                            # случай, когда выгрузка запущена на одной кассе
                            elif self.unloadTask.get('cashCode') and cash.code == self.unloadTask.get('cashCode'):
                                self.not_processed_cashes.append(cash.cashId)
                            else:
                                # случай выгрузки по магазинам, когда прогружаем все кассы
                                self.not_processed_cashes.append(cash.cashId)
                    if len(self.not_processed_cashes) == 0:
                        log.info('Не найдено касс для выгрузки продаж, задача %s будет удалена из очереди')
                        self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)
                    else:
                        log.info('Найдены кассы для выгрузки продаж: %s' % self.not_processed_cashes)
                else:
                    self.unloadTask = None
            except:
                self.unloadTask = None
        else:
            self.taskCountPass += 1
            log.info("Не удалось выполнить задачу %s для касс %s. Пробуем снова" % (self.unloadTask, self.not_processed_cashes))
        if self.unloadTask and len(self.not_processed_cashes) > 0:
            # собираем идентификаторы касс для выгрузки продаж
            self.all_cashes = len(self.not_processed_cashes)

            nodes_list = []  # данные для передачи в метод выгрузки продаж (shop, workDir, task, cash)
            for shop in filter(lambda x: x.id in self.notProcessedShops, nodes):
                for cash in filter(lambda x: x.cashId in self.not_processed_cashes, shop.cashes):
                    nodes_list.append((shop, self.workDirectory, self.unloadTask, cash))
            if self.multi_processing:
                log.debug('Запущена выгрузка в многопоточном режиме')
                # запускаем выгрузку для каждой кассы в отдельном процессе
                p = multiprocessing.Pool(self.processes_in_pool_count)
                # эта штука работает как-то так:
                # 1. создается набор процессов в кол-ве self.processes_in_pool_count
                # 2. каждый процесс берет задачу и выполняет
                # 3. когда задача выполнена, процесс отдает результат в callback(там делаем нашу обработку) и берет другую задачу, если они еще есть
                # 4. когда все задачи выполнены, процессы завершаются
                for i in range(len(nodes_list)):
                    p.apply_async(func=task_unload, args=nodes_list[i], callback=self.__post_unload_ext)
                p.close()
                p.join()
            else:
                for shop, work_dir, task, cash in nodes_list:
                    result = task_unload(shop=shop, work_directory=work_dir, task=task, cash=cash)
                    self.__post_unload_ext(result)

    def __post_unload_ext(self, unload_result):
        """
        Обработка результата выгрузки продаж, подсчет успешно/неуспешно прогруженных касс
        """
        res, cash_id = unload_result[0], unload_result[1]
        self.current_cash = cash_id
        log.debug('Текущая касса: %s' % self.current_cash)
        self.processed_cashes += 1
        if res:
            self.not_processed_cashes.remove(cash_id)
        log.info("Обработано касс: %s из %s" % (self.processed_cashes, self.all_cashes))
        # после того, как все кассы прогрузились, удаляем задачу из очереди
        if len(self.not_processed_cashes) == 0 and self.unloadTask:
            log.info("Сервис закончил выполнять задачу: %s" % self.unloadTask)
            self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)

    def __delete_task_from_queue(self, priority, task):
        """
        Удаление задачи из очереди
        """
        self.unloadTasks.task_done()  # нужно, чтобы у очереди изменился признак unfinished_tasks
        self.unloadTasks.queue.remove((priority, task))

    def __add_unload_task_to_queue(self):
        """
        Добавляет задачу в очередь
        """
        response.content_type = 'application/json'
        task = None
        try:
            task = json.loads(request.body.read(), encoding='utf-8')
            priority = task.pop('priority', 10)
            self.unloadTasks.put((priority, task), block=True, timeout=60)
            mess = u'Task %s added to queue' % task
            log.info(mess)
            result = {u'resultCode': 0, u'info': mess}
        except Queue.Full:
            mess = u'Queue is full. Can not add task %s to queue.' % task
            log.warning(mess)
            result = {u'resultCode': 10, u'info': mess}
        except Exception, e:
            log.warning(u'Ошибка REST сервиса. %s' % str(e))
            log.warning(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
            result = {u'resultCode': 100, u'info': str(e)}
        return json.dumps(result, ensure_ascii=False)

    def __get_unload_tasks_queue_status(self):
        """
        Возвращает таблицу с данными по задачам из очереди
        """
        try:
            queue_table = PrettyTable()
            queue_table.field_names = ['Индекс в очереди', 'Приоритет', 'Магазины/Кассы', 'Смены', 'Даты',
                                       'Тип выгрузки', 'Закрытые смены', 'Измененные смены', 'Прогресс', 'Текущий узел']
            # так называемые "стили" таблицы: выравнивание полей по левому краю
            # плюс ограничение длины для полей с магазинами и кассами
            queue_table.align = "l"
            queue_table._max_width = {queue_table.field_names[3]: 110}
            queue_table.hrules = ALL

            if len(self.unloadTasks.queue) > 0:
                for i in range(len(self.unloadTasks.queue)):
                    priority, task = self.unloadTasks.queue.__getitem__(i)
                    shops = task.get('shopId') if task.get('shopId') else ','.join(task.get('shopIds')) if (task.get('shopIds')) else None
                    cashes = task.get('cashCode') if task.get('cashCode') else ','.join(task.get('cashIds')) if (task.get('cashIds')) else None
                    progress = str("%d/%d" % (self.processed_cashes, self.all_cashes))
                    row = [i, priority, shops if shops else cashes, task.get('shifts'), task.get('dates'), task.get('granularity'),
                           task.get('closedShifts'), task.get('changedShifts'), progress, '%s' % self.current_cash]
                    queue_table.add_row(row)
            return queue_table.get_string()
        except Exception as e:
            log.warning('Ошибка получения состояния очереди: %s' % str(e))
            result = {u'resultCode': 500, u'info': str(e)}
            return json.dumps(result, ensure_ascii=False)
